# -*- coding: utf-8 -*-

# Copyright 2006-2012 Borislav Rizov <bobyATdclDOTbasDOTbg>, Department of Computational Linguistic, Institute for Bulgarian Language – Bulgarian Academy of Sciences
# 
# This file is part of Chooser.
# 
# Chooser is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Chooser is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Chooser.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

from common import *
from word import Word,  Words
from xml.dom import minidom
from hydra.wordnet import *
import re,  shutil

class Iterator:
    def __init__(self,  model):
        self.model = model
    def same(self,  current):
        return self.model.words[current].parent == self.model.current()
    def skip(self,  current):
        return self.passes(current) or self.same(current)
    def next(self):
        current = self.model._current + 1
        while current < len(self.model.words)  and self.skip(current):
            current += 1
        if current < len(self.model.words) :
            self.model._current = current
            return True
        return False
    def prev(self):
        current = self.model._current - 1
        while current >= 0 and self.skip(current):
            current -= 1
        if current >= 0:
            self.model._current = current
            return True
        return False

class All(Iterator):
    def passes(self,  current):
        return False

class NotChosen(Iterator):
    def passes(self,  current):
        return bool(self.model.words[current].selected())

class Ambiguous(Iterator):
    def passes(self,  current):
        return len(self.model.words[current].synsets()) < 2

class MatchWord(Iterator):
    def __init__(self,  model,  word=None,  case_sensitive = True):
        Iterator.__init__(self,  model)
        self.word = word
        self.case_sensitive = case_sensitive
    def passes(self,  current):
        if not self.word: word = self.model.current()
        else: word = self.word
        if not self.case_sensitive:
            flag = re.I | re.U
        else:
            flag = 0 | re.U
        if word.form():
            return not re.search(word.form(), self.model.words[current].form(),  flag)
        else:
            return not re.search(word.literal(),  self.model.words[current].literal(),  flag)

class MatchSense(Iterator):
    def __init__(self,  model,  word=None):
        self.word = word
        Iterator.__init__(self,  model)
    def passes(self,  current):
        current_sense = self.model.words[current].selected()
        sense = self.model.current().selected()
        return not current_sense or not sense or current_sense.ID() != sense.ID()


class Phantom(Iterator):
    def passes(self,  current):
        word = self.model.get_word(current)
        sel = word.selected()
#        print 'phantom',  sel,  bool(sel),  [x.ID() for x in word.senses()],  word.literal()
        res =  (not sel) or (sel.ID() in [x.ID() for x in word.senses()])
        if not res and sel:
            print "Phantom:",  sel.ID()
        return res

class Conjunction(Iterator):
    def __init__(self,  it1,  it2):
        Iterator.__init__(self,  it1.model)
        (self.it1,  self.it2) = it1,  it2
    def passes(self,  current):
        return self.it1.passes(current) or self.it2.passes(current)

class Model:
    FORWARD,  BACKWORD = range(2)

    def __init__(self, filename):
#		print filename 
        self._current = 0
        self.load(filename)
#		self.words = [Word(w + '1', w,  i) for i,  w in enumerate(['проба', 'word',  'cat',  'there',  'go',  'some',  'different',  'option'])]
#		self.direction = self.FORWARD

    def begin(self):
        self._current = 0
    def load(self,  filename):
        try:
            self.load_dom(filename)
        except:
            import traceback
            traceback.print_exc()
            try:
                self.load_chsr(filename)
            except:
                import traceback
                traceback.print_exc()
#				print 'chsr failed'
                raise "Unknown format"
        self.filename = filename
    def save(self, filename = None):
        if filename:
            self.filename = filename
        tmp_file = self.filename+'~'
        file = open(tmp_file, "w")
        dom = self.export_to_dom()
        file.write(dom.toxml().encode('utf8'))
        file.close()
        shutil.copy(tmp_file,  self.filename)
        #filme: add code for save
    def get_synset(self,  ili):
        synsets = wn.get("ili('%s')" % ili)
        for synset in synsets:
            if int(synset['lang']) == 1:
                return synset
    def load_chsr(self,  filename):
        self.words = []
        text = open(filename).read()
        parents = {}
        for tag,  m  in enumerate(re.findall(r"(\S*?){(.*?)}", text)):
#			if tag % 1 == 0: print tag
            w = m[0].decode('cp1251')
            info = re.match("(.*)#([^ ]+) ([^ ]+) ([^ ]+) (\d)",  m[1])
            l = info.group(1).decode('cp1251')
            ili = info.group(2)
            l = re.sub('#0', '', l)
            synset = self.get_synset(ili)
            next  = int(re.sub('\D', '',  info.group(4)))
#			print w,  l,  int(next),  tag
            word = Word(w,  l,  tag)
            if next != tag:
                if tag in parents:
                    parents[tag].add(word)
                else:
                    parents[tag] = Words(word)
                parents[next] = parents[tag]
            if synset:
                word.select(synset)
            self.words.append(word)
    def _split_parent(self, parent):
        parts = parent.split(':')
        if len(parts) == 2:
            return parts[0], int(parts[1])
        return parts[0], None
    def load_dom(self,  filename):
        self.words = []
        parents = {}
        dom = minidom.parse(filename)
        for i, word_el in enumerate(dom.getElementsByTagName('word')):
            w = word_el.getAttribute('w')
            l = word_el.getAttribute('l')
            l = re.sub('#0', '', l)

            p = word_el.getAttribute('p')
#			print 'p',  p
            s = word_el.getAttribute('s')
            e = word_el.getAttribute('e')
            pl = word_el.getAttribute('pl')
            word = Word(w,  l, i)
            if e:
                word.end = (e=='True')
            else:
                if re.search('[.!?]$', w): word.end=True
            if p:
                p, position = self._split_parent(p)
                if p in parents:
                    parents[p].add(word, position)
#                    print p, parents[p].words
                else:
                    parents[p] = Words(word, position)
#                    print p, parents[p].words
                    if e:
                        parents[p].end = (e=='True')
                    else:
                        if re.search('[.!?]$', w): word.end=(e=='True')
                stamped = parents[p]
                if pl:
                    stamped.lemma = pl
            else:
                stamped = word
            #set stamp
            stamped.user = word_el.getAttribute('u')
            stamped.time = word_el.getAttribute('t')
            
            if s:	word.select(Synset(long(s)))
            self.words.append(word)
#		print len(self.words)
        _current = dom.documentElement.getAttribute('current')
        if _current: self._current = min(len(self.words)-1,  int(_current))
    def export_to_dom(self):
        return self._export_to_dom(self.words,  self._current) 
    @classmethod
    def _encode_parent(self, word):
        return '%s:%s' % (id(word.parent), word.parent.words.index(word))
    @classmethod
    def _export_to_dom(self,  words,  current = None):
        dom = minidom.Document()
        root = dom.createElement('text')
        if current: root.setAttribute('current',  `current`)
        dom.appendChild(root)
        for word in words:
            word_el = dom.createElement('word')
            word_el.setAttribute('w',  word.word)
            word_el.setAttribute('l',  word.lemma)
            if word.parent:
                word_el.setAttribute('p',  Model._encode_parent(word))
                if word.parent.lemma and word == word.parent.words[0]:
                    word_el.setAttribute('pl',  word.parent.lemma)
#                    print 'pl',  word.parent.lemma
                if word == word.parent.words[0]:
                    if word.parent.end!=None: word_el.setAttribute('e', str(word.parent.end))
            else:
                if word.end!=None: word_el.setAttribute('e',  str(word.end))

            if word.selected():#
                word_el.setAttribute('s',  str(word.selected().ID()))
                stamp = word.stamp()
                word_el.setAttribute('u',  str(stamp[0]))
                word_el.setAttribute('t',  str(stamp[1]))
            root.appendChild(word_el)
        return dom
    def select(self,  w,  multi):
        if multi:
            if w.parent == None:
                word = self.words[self._current]
                if word.parent:
                    word.parent.add(w)
                else:
                    parent = Words(word)
                    parent.add(w)
        else:
#			print type(w.tag),  w.tag
            self._current = w.index
    def deselect(self, w, multi):
        if multi:
            if w.parent:
                sel = w.parent.remove(w)	
                self._current = sel.index
#				print self.current().literal()
    def is_selected(self,  w):
        word = self.words[self._current]
        if word.parent:
            return w in word.parent.words
        else:
            return w == word
    def move(self,  direction):
        if direction == self.FORWARD:
            return self.next()
        else:
            return self.prev()

    def find(self, word,  case_sensitive,  direction):
        it = self.iterator
        
        self.iterator = Conjunction(All(self),  MatchWord(self,  word,  case_sensitive))
        res = self.move(direction)
        
        self.iterator = it
        return res
    def set_iterator(self,  IteratorClass,  iterate_current=False,  iterate_sense=False,   *args):
        self.iterator = IteratorClass(self,  *args)
        if iterate_current:
            self.iterator = Conjunction(self.iterator,  MatchWord(self))
        if iterate_sense:
            self.iterator = Conjunction(self.iterator,  MatchSense(self))
        
    def simple_current(self):
        return self.words[self._current]
    def get_word(self,  i):
        word = self.words[i]
        if word.parent:
            return word.parent
        else:
            return word
    def current(self):
#        print 'current',  self._current
        return self.get_word(self._current)
    def is_end(self,  word):
        if word.parent:
            if word.parent.end:
                for w in word.parent.words:
                    if w.tag > word.tag:
                        return False
                return True
        else:
            return word.end
    def next(self):
        return self.iterator.next()
    def prev(self):
        return self.iterator.prev()
    def insert(self,  form,  lemma):
        word = Word(form,  lemma,  self._current)
        for i in xrange(self._current,  len(self.words)):
#			print 'i',  i
            self.words[i].index = i + 1
        self.words.insert(self._current,  word)
        return word
    def simple_delete(self):
        del self.words[self._current]
        for i in xrange(self._current,  len(self.words)):
            self.words[i].index = i
        if self._current>= len(self.words):
            self._current = len(self.words) - 1
    def delete(self):
        if isinstance(self.current(),  Words):
            for word in self.current().words:
#				print word.word
                self._current = word.index
                self.simple_delete()
        else:
            self.simple_delete()
    def append(self,  form,  lemma):
        self._current = len(self.words)
        self.insert(form,  lemma)
    def stats(self):
        stats = {}
        parents = {}
        stats['selected'] = 0
        stats['single words'] = 0
        stats['single selected'] = 0
        for word in self.words:
            if word.parent:
                parents[word.parent] = word.selected()
            else:
                stats['single words'] += 1
                if word.selected():
                    stats['single selected'] += 1
        stats['multi words'] = len(parents)
        stats['multi selected'] = len([1 for p in parents if parents[p]])
        stats['words'] = len(self.words)
        return stats
    
def text_value(node):
        for child in node.childNodes:
            if child.nodeType == child.TEXT_NODE:
                return child.nodeValue

class Node:
    def __init__(self, node):
        self.node = node
    @classmethod
    def create(cl, name, value, path):
        node = Element(name)
        x = Text()
        x.data = value
        node.appendChild(x)
        node.setAttribute('PATH', path)
        return Node(node)
    def value2(self):
        pos = self.pos()
        if pos:
            return self.value() + " " + pos
        return self.value()
    def value(self):
        txt = text_value(self.node)
        if not txt or txt == '':
            return "<empty>"
        return text_value(self.node)
    def path(self, value = None):
        if value:
            self.set_attribute('PATH', value)
        else:
            return self.attribute('PATH')
    def add_arc(self,  dest,  arc,  links):
        anode = self.create_element('Node')
        anode.set_attribute('Dest',  `dest`)
        anode.set_attribute('Arc',  arc)
        anode = links.append_child(anode)
    def get_element_by_name(self,  name):
        l = self.node.getElementsByTagName(name)
        if l and len(l) > 0:
            return Node(l[0])
        else:
            return None
    def link(self, dest,  arc):
        links = self.get_element_by_name('Links')
        if links:
#			links = Node(l[0])
            for child in links.children():
                if dest == int(child.attribute('Dest')):
                    child.set_attribute('Arc',  arc)
                    return
            self.add_arc(dest,  arc,  links)
        else:
            links = self.create_element('Links')
            links = self.append_child(links)
            self.add_arc(dest,  arc,  links)
    def append_child(self,  node):
        return Node(self.node.appendChild(node.node))
    def  create_element(self,  name):
        return Node(self.node.ownerDocument.createElement(name))
    def del_link(self,  dest):
        links = self.get_element_by_name('Links')
        if links:
            for child in links.children():
                if dest == int(child.attribute('Dest')):
                    child.destruct()
    def links(self):
        result = dict()
        links = self.get_element_by_name('Links')
        if links:
            for child in links.children():
                dest = child.attribute('Dest')
                arc = child.attribute('Arc')
                result[int(dest)] = arc.encode('utf8')
        return result
    def pos(self):
        return self.attribute('POS')
    def set_pos(self,  pos):
        return self.set_attribute('POS',  pos)
    def name(self):
        return self.node.nodeName
    def destruct(self):
        parent = self.node.parentNode
        parent.removeChild(self.node).unlink()
        self.node = None
    def insert_before(self, node):
        parent = self.node.parentNode
        parent.insertBefore(node.node, self.node)
    def children(self):
        return [Node(n) for n in self.node.childNodes if n.nodeType == n.ELEMENT_NODE]
    def parent(self):
        return Node(self.node.parentNode)
    def __eq__(self, node):
        return self.node.isSameNode(node.node)
    def attribute(self,  name):
        return self.node.getAttribute(name)
    def set_attribute(self,  name,  value):
        self.node.setAttribute(name,  value)
    def remove_attribute(self,  name):
        self.node.removeAttribute(name)
class Sentence(Node):
    def __init__(self, node):
        Node.__init__(self, node)
    def node_value_pos(self, node):
        value = node.value()
        pos = node.pos()
        if pos:
            value += '\\n' + pos
        #print "name", node.name()
        if node.name() in ('XP', 'P'):
            return value
        else:
            return " " + value
    def _node_value(self, node):
        value = node.value()
        #print "name", node.name()
        if node.name() in ('XP', 'P'):
            return value
        else:
            return " " + value
    def to_string(self):
        return "".join([self._node_value(n) for n in self.children() if n.name() != 'XP'])
    def path(self):
        return "0"
